#include <iostream>
#include <arpa/inet.h>
#include <cstdlib>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <memory.h>
#include <unistd.h>
#include <time.h>

#include "discover_lib.h"

using namespace std;

/**********************************************************************
 * summary : Constructer (prepare socket)
 * return  : -
 *********************************************************************/
CDiscover::CDiscover()
{
	s_sock = -1;
	r_sock = -1;

	PrepareBroadcastAddress();

	if(!SetupRecvSocket()){
		cerr << "failed to prepare socket" << endl;
		exit(-1);
	}
}
/**********************************************************************
 * summary : Setup send socket 
 * return  : -
 *********************************************************************/
in_addr_t CDiscover::GetIPAddr(const char *device)
{
	struct ifreq ifr;
	int sock;

	if((sock = ::socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		return 0;
	
	ifr.ifr_addr.sa_family = AF_INET;
	strncpy(ifr.ifr_name, device, IFNAMSIZ - 1);

	if(ioctl(sock, SIOCGIFADDR, &ifr) < 0)
		return 0;

	close(sock);

	return inet_addr(inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr));
}
/**********************************************************************
 * summary : Setup socket 
 * return  : -
 *********************************************************************/
int CDiscover::CreateINETSocket(const char *device)
{
	////
	// Set up the address we're going to bind to.
	////
	struct sockaddr_in name;
	int sock;

	name.sin_family = AF_INET;
	name.sin_port = htons(CLIENT_PORT);
	if(device)
		name.sin_addr.s_addr = GetIPAddr(device);
	else
		name.sin_addr.s_addr = INADDR_ANY;
	memset(name.sin_zero, 0, sizeof (name.sin_zero));

	// Make a socket...
	if((sock = ::socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
		cerr << "Unable to make socket" << endl;
		return -1;
	}

	// Set the REUSEADDR option so that we don't fail to start
	int flag = 1;
	if(::setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
			(char *)&flag, sizeof flag) < 0){
		cerr << "Can't set SO_REUSEADDR option" << endl;
		return -1;
	}

	// Set the BROADCAST option so that we can broadcast DHCP responses.
	if(::setsockopt(sock, SOL_SOCKET, SO_BROADCAST,
            (char *)&flag, sizeof flag) < 0){
		cerr << "Can't set SO_BROADCAST option" << endl;
		return -1;
	}

	// Bind the socket to this interface's IP address.
	if(::bind(sock, (struct sockaddr *)&name, sizeof name) < 0){
		cerr << "Bind error" << endl;
		return -1;
	}

	return sock;
}
/**********************************************************************
 * summary : Prepare broadcast address
 * return  : -
 *********************************************************************/
void CDiscover::PrepareBroadcastAddress()
{
	memset((char *)&m_sockaddr_broadcast, 0, sizeof(struct sockaddr_in));
	m_sockaddr_broadcast.sin_family = AF_INET;
	m_sockaddr_broadcast.sin_port = htons(SERVER_PORT);
	m_sockaddr_broadcast.sin_addr.s_addr = INADDR_BROADCAST;
}
/**********************************************************************
 * summary : Setup send socket 
 * return  : -
 *********************************************************************/
bool CDiscover::SetupSendSocket(const char *device)
{
	if(s_sock >= 0){
		close(s_sock);
		s_sock = -1;
	}

	if((s_sock = CreateINETSocket(device)) < 0)
		return false;

	return true;
}
/**********************************************************************
 * summary : Setup recv socket 
 * return  : -
 *********************************************************************/
bool CDiscover::SetupRecvSocket()
{
	if((r_sock = CreateINETSocket(NULL)) < 0)
		return false;

	return true;
}

/**********************************************************************
 * summary : send indicated packet
 * return  : -
 *********************************************************************/
CDiscover::~CDiscover()
{
	if(s_sock >= 0){
		close(s_sock);
	}

	if(r_sock >= 0){
		close(r_sock);
	}
}

/**********************************************************************
 * summary : Make new transaction ID
 * return  : transaction ID
 *********************************************************************/
uint32_t CDiscover::GetNewTransactionID() const
{
	while(1){
		srand((unsigned)time(NULL));
		unsigned long num = (unsigned long)rand();
		if(num){
			return num;
		}
	}
}

/**********************************************************************
 * summary : send indicated packet
 * return  : -
 *********************************************************************/
bool CDiscover::SendPacket(const discover_packet_s* packet, const char *device)
{

	if(!SetupSendSocket(device)){
		cerr << "failed to prepare socket" << endl;
		return -1;
	}

	int send = ::sendto(
						s_sock,
						(char*)packet,
						sizeof(discover_packet_s),
						0,
						(struct sockaddr*)&m_sockaddr_broadcast,
						sizeof(m_sockaddr_broadcast)
					);
	return (send == sizeof(discover_packet_s));
}

/**********************************************************************
 * summary : Broadcast find packet
 * return  : transaction id
 *********************************************************************/
uint32_t CDiscover::SendFindPacket(const char* device)
{
	discover_packet_s packet;

	memset(&packet, 0, sizeof(discover_packet_s));
	packet.message_type = htonl(DISCOVER_REQ);
	packet.transaction_id = GetNewTransactionID();

	if(SendPacket(&packet, device))
		return packet.transaction_id;
	
	return 0;
}

/**********************************************************************
 * summary : Broadcast config request packet
 * return  : transaction id
 *********************************************************************/
uint32_t CDiscover::SendConfigPacket(
	const char* macaddr,
	const network_info_s* networkInfo,
	const char* device
)
{
	discover_packet_s packet;

	memset(&packet, 0, sizeof(discover_packet_s));
	packet.message_type = htonl(IP_OFFER_REQ);
	packet.transaction_id = GetNewTransactionID();
	memcpy(&packet.mac_addr, macaddr, 6);
	memcpy(&packet.net_info, networkInfo, sizeof(network_info_s));

	if(SendPacket(&packet, device)){
		return packet.transaction_id;
	}
	
	return 0;
}

/**********************************************************************
 * summary : Receive Packet
 * return  : -
 *********************************************************************/
bool CDiscover::RecvPacket(
	discover_packet_s* packet
)
{
	struct sockaddr sender;
	socklen_t len = sizeof(sender);
	int size = ::recvfrom(
						r_sock,
						(char*)packet,
						sizeof(discover_packet_s),
						0,
						&sender,
						&len
					);
	return size == sizeof(discover_packet_s);
}

/**********************************************************************
 * summary : Receive packet until timeout
 * return  : -
 *********************************************************************/
bool CDiscover::RecvReplyPacket(
	discover_packet_s* packet,
	uint32_t transactionID,
	unsigned int timeout
)
{
	struct timeval* tout = NULL;
	if(timeout){
		tout = new struct timeval;
		tout->tv_sec = timeout;
		tout->tv_usec = 0;
	}
		
	while(1){
		fd_set readfds;
		FD_ZERO(&readfds);
		FD_SET(r_sock, &readfds);

		switch(select(FD_SETSIZE, &readfds, NULL, NULL, tout)){
		case 0:
			// timeout
		case -1:
			// error;
			return false;
		default:
			break;
		}

		if(!FD_ISSET(r_sock, &readfds)){
			cerr << "FD is not set" << endl;
			return false;
		}

		if(!RecvPacket(packet)){
			return false;
		}

		if(packet->transaction_id == transactionID){
			return true;
		}
	}
}

/**********************************************************************
 * summary : get list of interface
 * return  : number of interface
 *********************************************************************/
int32_t CDiscover::GetInterfaceList(struct ifreq **ifr)
{
	struct ifconf ifc;
	int nr_ifs;
	int max_ifs = 10;
	int sock;

	sock = socket(AF_INET, SOCK_DGRAM, 0);

	while(1){
		ifc.ifc_len = sizeof(struct ifreq) * max_ifs;

		*ifr = new ifreq[max_ifs];
		ifc.ifc_buf = (char *)*ifr;

		if(ioctl(sock, SIOCGIFCONF, &ifc) < 0){
			delete [] *ifr;
			return -1;
		}
		
		if(ifc.ifc_len == (int32_t)(sizeof(struct ifreq) * max_ifs)){
			max_ifs += 10;
			continue;
		}

		break;
	}

	nr_ifs = ifc.ifc_len / sizeof(struct ifreq);

	return nr_ifs;
}

/**********************************************************************
 * summary : get list of interface
 * return  : number of interface
 *********************************************************************/
void CDiscover::FreeInterfaceList(struct ifreq *ifr)
{
	if(ifr != NULL){
		delete [] ifr;
	}
}
